#include <linux/init.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/kthread.h>
#include <linux/mutex.h>
#include <linux/delay.h>
#include <linux/list.h>
#include <asm-generic/uaccess.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Britton");
MODULE_DESCRIPTION("Simple module featuring list access");

struct list_item {
	struct list_head list;
	int value;
};

#define KFLAGS (__GFP_WAIT | __GFP_IO | __GFP_FS) 

#define ENTRY_NAME "list"
#define PERMS 0644
#define PARENT NULL
static struct file_operations fops;

static struct task_struct *producer;
static struct task_struct *mover;
static struct task_struct *consumer;

static struct list_head list_1;
static struct list_head list_2;

static struct mutex list_1_mutex;
static struct mutex list_2_mutex;

static char *message;
static int read_p;





int producer_run(void *data) {
	const int sleep_time = 1;
	const int num_to_add = 2;
	int start_value = 1;
	int i;
	struct list_item *item;

	while (!kthread_should_stop()) {
		ssleep(sleep_time);

		mutex_lock_interruptible(&list_1_mutex);
		for (i = 0; i < start_value % num_to_add; i+=1) {
			item = kmalloc(sizeof(struct list_item), KFLAGS);
			item->value = start_value;
			list_add_tail(&item->list, &list_1);
		}
		mutex_unlock(&list_1_mutex);

		start_value += sleep_time;
	}
	return start_value;
}

int mover_run(void *data) {
	struct list_item *item;
	const int sleep_time = 2;
	const int num_to_add = 5;
	int start_value = 1;
	int i;

	while (!kthread_should_stop()) {
		ssleep(sleep_time);

		mutex_lock_interruptible(&list_1_mutex);
		mutex_lock_interruptible(&list_2_mutex);
		for (i = 0; i < start_value % num_to_add; i+=1) {
			if (!list_empty(&list_1)) {
				item = list_first_entry(&list_1, struct list_item, list);
				list_move_tail(&item->list, &list_2);
			}
		}
		mutex_unlock(&list_2_mutex);
		mutex_unlock(&list_1_mutex);

		start_value += sleep_time;
	}
	return start_value;
}

int consumer_run(void *data) {
	struct list_item *item;
	const int sleep_time = 3;
	const int num_to_add = 7;
	int start_value = 1;
	int i;

	while (!kthread_should_stop()) {
		ssleep(sleep_time);

		mutex_lock_interruptible(&list_2_mutex);
		for (i = 0; i < start_value % num_to_add; i+=1) {
			if (!list_empty(&list_2)) {
				item = list_first_entry(&list_2, struct list_item, list);
				list_del_init(&item->list);
				kfree(item);
			}
		}
		mutex_unlock(&list_2_mutex);

		start_value += sleep_time;
	}
	return start_value;
}





int list_proc_open(struct inode *sp_inode, struct file *sp_file) {
	int num_entries = 0;
	int first_p;
	struct list_head *ptr;
	struct list_item *item;
	char buf[10];

	printk("proc called open\n");

	mutex_lock_interruptible(&list_1_mutex);
	list_for_each(ptr, &list_1) {
		num_entries += 1;
	}
	mutex_unlock(&list_1_mutex);

	mutex_lock_interruptible(&list_2_mutex);
	list_for_each(ptr, &list_2) {
		num_entries += 1;
	}
	mutex_unlock(&list_2_mutex);

	read_p = 1;
	message = kmalloc(sizeof(char) * (50 + (3 * num_entries)), KFLAGS);
	if (message == NULL) {
		printk("ERROR, counter_proc_open");
		return -ENOMEM;
	}
	strcpy(message, "");

	mutex_lock_interruptible(&list_1_mutex);
	first_p = 1;
	strcat(message, "Produced items:\n");
	list_for_each(ptr, &list_1) {
		item = list_entry(ptr, struct list_item, list);
		if (first_p) {
			first_p = 0;
			sprintf(buf, "%d", item->value);
			strcat(message, buf);
		}
		else {
			sprintf(buf, ", %d", item->value);
			strcat(message, buf);
		}
	}
	strcat(message, "\n");
	mutex_unlock(&list_1_mutex);

	mutex_lock_interruptible(&list_2_mutex);
	first_p = 1;
	strcat(message, "Moved items:\n");
	list_for_each(ptr, &list_2) {
		item = list_entry(ptr, struct list_item, list);
		if (first_p) {
			first_p = 0;
			sprintf(buf, "%d", item->value);
			strcat(message, buf);
		}
		else {
			sprintf(buf, ", %d", item->value);
			strcat(message, buf);
		}
	}
	strcat(message, "\n");
	mutex_unlock(&list_2_mutex);
	return 0;
}

ssize_t list_proc_read(struct file *sp_file, char __user *buf, size_t size, loff_t *offset) {
	int len;
	printk("proc called read\n");
	
	read_p = !read_p;
	if (read_p) {
		return 0;
	}
	
	len = strlen(message);
	copy_to_user(buf, message, len);
	return len;
}

int list_proc_release(struct inode *sp_inode, struct file *sp_file) {
	printk("proc called release\n");
	kfree(message);
	return 0;
}





static int compete_init(void) {
	printk("/proc/%s create\n", ENTRY_NAME); 
	
	mutex_init(&list_1_mutex);
	mutex_init(&list_2_mutex);

	INIT_LIST_HEAD(&list_1);
	INIT_LIST_HEAD(&list_2);

	producer = kthread_run(producer_run, NULL, "producer thread"); 
	if (IS_ERR(producer)) {
		printk("ERROR! kthread_run, producer thread\n");
		return PTR_ERR(producer);
	}

	mover = kthread_run(mover_run, NULL, "mover thread");
	if (IS_ERR(mover)) {
		printk("ERROR! kthread_run, mover thread\n");
		return PTR_ERR(mover);
	}
	
	consumer = kthread_run(consumer_run, NULL, "consumer thread");
	if (IS_ERR(consumer)) {
		printk("ERROR! kthread_run, consumer thread\n");
		return PTR_ERR(consumer);
	}

	fops.open = list_proc_open;
	fops.read = list_proc_read;
	fops.release = list_proc_release;

	if (!proc_create(ENTRY_NAME, PERMS, NULL, &fops)) {
		printk("ERROR! proc_create\n");
		remove_proc_entry(ENTRY_NAME, NULL);
		return -ENOMEM;
	}
	return 0;
}

static void compete_exit(void) {
	int ret;
	struct list_head *temp, *ptr;
	struct list_item *item;
	printk("Removing module\n");
	
	ret = kthread_stop(producer);
	if (ret != -EINTR)
		printk("producer thread has stopped\n");

	ret = kthread_stop(mover);
	if (ret != -EINTR)
		printk("mover thread has stopped\n");
	
	ret = kthread_stop(consumer);
	if (ret != -EINTR)
		printk("consumer thread has stopped\n");

	list_for_each_safe(ptr, temp, &list_1) {
		item = list_entry(ptr, struct list_item, list);
		list_del(ptr);
		kfree(item);
	}
	list_for_each_safe(ptr, temp, &list_2) {
		item = list_entry(ptr, struct list_item, list);
		list_del(ptr);
		kfree(item);
	}

	printk("Removing /proc/%s.\n", ENTRY_NAME);
	remove_proc_entry(ENTRY_NAME, NULL);
}

module_init(compete_init);
module_exit(compete_exit);
